import asyncio
from typing import TYPE_CHECKING, Any, List, Optional
from ain.types import (
    SetOperation,
    GetOperation,
    SetOperationType,
    GetOperationType,
    SetMultiOperationType,
    SetMultiOperation,
    TransactionInput,
    ValueOnlyTransactionInput,
    SetMultiTransactionInput,
    EvalRuleInput,
    EvalOwnerInput,
    MatchInput,
)
from ain.ain_db.push_id import PushId

if TYPE_CHECKING:
    from ain.ain import Ain

class Reference:
    _path: str
    _key: Optional[str]
    _ain: "Ain"
    _pushId: PushId
    _isGlobal: bool

    def __init__(self, ain: "Ain", path: str = None):
        self._path = Reference.sanitizeRef(path)
        pathArr = self._path.split("/")
        if len(pathArr) > 0:
            self._key = pathArr[-1]
        else:
            self._key = None
        self._ain = ain
        self._pushId = PushId()
        self._isGlobal = False

    @property
    def path(self) -> str:
        return self._path

    @property
    def key(self) -> Optional[str]:
        return self._key

    def setIsGlobal(self, isGlobal):
        """
        Sets the global path flag.
        """
        self._isGlobal = isGlobal

    def push(self) -> "Reference":
        """
        Creates a key for a new child but doesn't set any values.
        """
        newKey = "/" + self._pushId.generate()
        return Reference(self._ain, Reference.extendPath(self._path, newKey))

    async def pushValue(self, value: Any = None) -> Any:
        """
        Sets the value at a new child of self.path;
        """
        return await self.push().setValue(ValueOnlyTransactionInput(value=value))

    async def getValue(self, path: str = None) -> Any:
        """
        Returns the value at the path.
        """
        req = self.buildGetRequest("GET_VALUE", Reference.extendPath(self.path, path))
        return await self._ain.provider.send("ain_get", req)

    async def getRule(self, path: str = None) -> Any:
        """
        Returns the rule at the path.
        """
        req = self.buildGetRequest("GET_RULE", Reference.extendPath(self.path, path))
        return await self._ain.provider.send("ain_get", req)

    async def getOwner(self, path: str = None) -> Any:
        """
        Returns the owner config at the path.
        """
        req = self.buildGetRequest("GET_OWNER", Reference.extendPath(self.path, path))
        return await self._ain.provider.send("ain_get", req)

    async def getFunction(self, path: str = None) -> Any:
        """
        Returns the function config at the path.
        """
        req = self.buildGetRequest("GET_FUNCTION", Reference.extendPath(self.path, path))
        return self._ain.provider.send("ain_get", req)

    async def get(self, gets: List[GetOperation]) -> Any:
        """
        Returns the value / write rule / owner rule / function hash at multiple paths.

        Args:
            gets (List[GetOperation]): Array of get requests
                Could be any one from "VALUE", "RULE", "OWNER", "FUNC" or a combination of them as an array.
        """
        extendedGets: List[GetOperation] = []
        for get in gets:
            extendedGets.append(
                GetOperation(
                    type=get.type,
                    ref=Reference.extendPath(self._path, getattr(get, "ref", None)),
                    is_global=getattr(get, "is_global", None),
                )
            )
        req = {"type": "GET", "op_list": extendedGets}
        return await self._ain.provider.send("ain_get", req)

    async def deleteValue(self, transactionInput: ValueOnlyTransactionInput = None) -> Any:
        """
        Deletes a value.

        Args:
            transactionInput (ValueOnlyTransactionInput):
                A transaction input object.
                Any value given will be overwritten with null.
        """
        txInput = ValueOnlyTransactionInput()
        if transactionInput is not None:
            txInput = transactionInput
        txInput.value = None
        ref = Reference.extendPath(self._path, getattr(txInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                txInput,
                ref,
                "SET_VALUE",
                self._isGlobal
            )
        )

    async def setFunction(self, transactionInput: ValueOnlyTransactionInput) -> Any:
        """
        Sets a function config.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_FUNCTION",
                self._isGlobal
            )
        )

    async def setOwner(self, transactionInput: ValueOnlyTransactionInput) -> Any:
        """
        Sets the owner rule.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_OWNER",
                self._isGlobal
            )
        )

    async def setRule(self, transactionInput: ValueOnlyTransactionInput) -> Any:
        """
        Sets the write rule.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_RULE",
                self._isGlobal
            )
        )

    async def setValue(self, transactionInput: ValueOnlyTransactionInput) -> Any:
        """
        Sets a value.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_VALUE",
                self._isGlobal
            )
        )

    async def incrementValue(self, transactionInput: ValueOnlyTransactionInput) -> Any:
        """
        Increments the value.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "INC_VALUE",
                self._isGlobal
            )
        )

    async def decrementValue(self, transactionInput: ValueOnlyTransactionInput) -> Any:
        """
        Decrements the value.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "DEC_VALUE",
                self._isGlobal
            )
        )

    async def set(self, transactionInput: SetMultiTransactionInput) -> Any:
        """
        Processes multiple set operations.
        """
        return await self._ain.sendTransaction(
            Reference.extendSetMultiTransactionInput(
                transactionInput,
                self._path,
            )
        )

    async def evalRule(self, params: EvalRuleInput) -> bool:
        """
        Returns the rule evaluation result. True if the params satisfy the write rule,
        false if not.
        """
        address = self._ain.wallet.getImpliedAddress(getattr(params, "address", None))
        req = {
            "address": address,
            "ref": Reference.extendPath(self._path, getattr(params, "ref", None)),
            "value": params.value,
        }
        if hasattr(params, "timestamp"):
            req["timestamp"] = params.timestamp
        return await self._ain.provider.send("ain_evalRule", req)

    async def evalOwner(self, params: EvalOwnerInput) -> Any:
        """
        Returns the owner evaluation result.
        """
        address = self._ain.wallet.getImpliedAddress(getattr(params, "address", None))
        req = {
            "address": address,
            "ref": Reference.extendPath(self._path, getattr(params, "ref", None)),
            "permission": params.permission,
        }
        return await self._ain.provider.send("ain_evalOwner", req)

    async def matchFunction(self, params: MatchInput = None) -> Any:
        """
        Returns the function configs that are related to the input ref.
        """
        ref = self._path
        if params is not None and hasattr(params, "ref"):
            ref = Reference.extendPath(ref, getattr(params, "ref", None))
        return await self._ain.provider.send("ain_matchFunction", {"ref": ref})

    async def matchRule(self, params: MatchInput = None) -> Any:
        """
        Returns the rule configs that are related to the input ref.
        """
        ref = self._path
        if params is not None and hasattr(params, "ref"):
            ref = Reference.extendPath(ref, getattr(params, "ref", None))
        return await self._ain.provider.send("ain_matchRule", {"ref": ref})

    async def matchOwner(self, params: MatchInput = None) -> Any:
        """
        Returns the owner configs that are related to the input ref.
        """
        ref = self._path
        if params is not None and hasattr(params, "ref"):
            ref = Reference.extendPath(ref, getattr(params, "ref", None))
        return await self._ain.provider.send("ain_matchOwner", {"ref": ref})

    @staticmethod
    def buildGetRequest(type: GetOperationType, ref: str) -> dict:
        """
        Builds a get request.
        """
        return {"type": type, "ref": Reference.sanitizeRef(ref)}

    @staticmethod
    def extendPath(basePath: str = None, extension: str = None) -> str:
        """
        Returns a path that is the basePath extended with extension.
        """
        sanitizedBase = Reference.sanitizeRef(basePath)
        sanitizedExt = Reference.sanitizeRef(extension)
        if sanitizedBase == "/":
            return sanitizedExt
        if sanitizedExt == "/":
            return sanitizedBase
        return sanitizedBase + sanitizedExt

    @staticmethod
    def extendSetTransactionInput(
        input: ValueOnlyTransactionInput,
        ref: str,
        type: SetOperationType, 
        isGlobal: bool
    ) -> TransactionInput:
        """
        Decorates a transaction input with an appropriate type, ref and value.

        Args:
            input (ValueOnlyTransactionInput): A transaction input object
            ref (str): The path at which set operations will take place
            type (SetOperationType): A type of set operations
        """
        operation = SetOperation(
            type=type,
            ref=ref,
            value=getattr(input, "value", None),
            is_global=getattr(input, "is_global", isGlobal),
        )
        return TransactionInput(
            operation=operation,
            parent_tx_hash=getattr(input, "parent_tx_hash", None),
        )

    @staticmethod
    def extendSetMultiTransactionInput(
        input: SetMultiTransactionInput,
        ref: str,
        type: SetMultiOperationType = "SET",
    ):
        """
        Decorates a transaction input with an appropriate type and op_list.

        Args:
            input (SetMultiTransactionInput): A transaction input object
            ref (str): The path at which set operations will take place
            type (SetMultiOperationType): A type of set operations
        """
        op_list: List[SetOperation] = []
        for op in input.op_list:
            op_list.append(
                SetOperation(
                    type=op.type,
                    ref=Reference.extendPath(ref, op.ref),
                    value=getattr(op, "value", None),
                    is_global=getattr(op, "is_global", None)
                )
            )
            
        operatiron = SetMultiOperation(type, op_list)
        return TransactionInput(
            operation=operatiron,
            parent_tx_hash=getattr(input, "parent_tx_hash", None),
            nonce=getattr(input, "nonce", None),
            address=getattr(input, "address", None),
            timestamp=getattr(input, "timestamp", None),
        )

    @staticmethod
    def sanitizeRef(ref: str = None):
        """
        Returns a sanitized ref. If should have a slash at the
        beginning and no slash at the end.
        """
        if ref is None:
            return "/"
        return "/" + "/".join(filter(lambda s: s != "", ref.split("/")))



from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from ain.types import (
    TransactionBody,
    ValueOnlyTransactionInput,
    ECDSASignature,
)
from ain.utils import (
    toBytes,
    toChecksumAddress,
    bytesToHex,
    pubToAddress,
    privateToPublic,
    privateToAddress,
    ecSignMessage,
    ecSignTransaction,
    ecRecoverPub,
    ecVerifySig,
)

if TYPE_CHECKING:
    from ain.ain import Ain

class Account:
    private_key: str
    public_key: str
    address: str
    
    def __init__(self, privateKey: Union[bytes, str] = None):
        if type(privateKey) is bytes:
            privateKeyBytes = privateKey
        elif type(privateKey) is str:
            privateKeyBytes = bytes.fromhex(privateKey)
        else:
            raise TypeError('privateKey has invalid type')

        self.private_key = privateKeyBytes.hex()
        self.public_key = privateToPublic(privateKeyBytes).hex()
        self.address = privateToAddress(privateKeyBytes)

    def __str__(self):
        return "\n".join(
            (
                "{",
                f"  address: '{self.address}'",
                f"  private: '{self.private_key}'",
                f"  public: '{self.public_key}'",
                "}",
            )
        )

Accounts = Dict[str, Account]

class Wallet:
    defaultAccount: Optional[Account]
    accounts: Accounts
    ain: "Ain"
    chainId: int

    def __init__(self, ain: "Ain", chainId: int):
        self.defaultAccount = None
        self.accounts: Accounts = {}
        self.ain = ain
        self.chainId = chainId

    @property
    def length(self) -> int:
        return len(self.accounts)

    def getPublicKey(self, address: str) -> str:
        """
        Returns the full public key of the given address.
        """
        checksummed = toChecksumAddress(address)
        if checksummed not in self.accounts:
            return ''
        return self.accounts[checksummed].public_key

    # TODO(kriii): implement this function.
    def create(self):
        """
        Creates {numberOfAccounts} new accounts and add them to the wallet.
        """
        pass
    
    def isAdded(self, address: str) -> bool:
        """
        Returns whether the address has already been added to the wallet.
        """
        return address in self.accounts

    def add(self, privateKey: str) -> str:
        """
        Adds a new account from the given private key.
        """
        newAccount = Account(privateKey)
        self.accounts[newAccount.address] = newAccount
        return newAccount.address

    def addAndSetDefaultAccount(self, privateKey: str) -> str:
        """
        Adds a new account from the given private key and sets the new account as the default account.
        """
        address = self.add(privateKey)
        self.setDefaultAccount(address)
        return address

    # TODO(kriii): implement this function.
    def addFromHDWallet(self):
        """
        Adds an account from a seed phrase. Only the account at the given
        index (default = 0) will be added.
        """
        pass

    # TODO(kriii): implement this function.
    def addFromV3Keystore(self):
        """
        Adds an account from a V3 Keystore.
        """
        pass
    
    def remove(self, address: str):
        """
        Removes an account
        """
        addressToRemove = toChecksumAddress(address)
        if addressToRemove not in self.accounts:
            raise ValueError("Can't find account to remove")
        del self.accounts[addressToRemove]
        if self.defaultAccount is not None and self.defaultAccount.address == addressToRemove:
            self.removeDefaultAccount()

    def setDefaultAccount(self, address: str):
        """
        Sets the default account as {address}. The account should be already added
        in the wallet.
        """
        checksumed = toChecksumAddress(address)
        if checksumed not in self.accounts:
            raise ValueError("Add the account first before setting it to defaultAccount.")
        self.defaultAccount = self.accounts[checksumed]

    def removeDefaultAccount(self):
        """
        Removes a default account (sets it to None).
        """
        self.defaultAccount = None

    def clear(self):
        """
        Clears the wallet (remove all account information).
        """
        self.accounts.clear()
        self.removeDefaultAccount()

    def getImpliedAddress(self, inputAddress= None) -> str:
        """
        Returns the "implied" address. If address is not given,
        it returns the defaultAccount. It throws an error if
        an address is not given and defaultAccount is not set, or
        the specified address is not added to the wallet.
        """
        if inputAddress is not None:
            address = inputAddress
        elif self.defaultAccount is not None:
            address = self.defaultAccount.address
        else:
            raise ValueError("You need to specify the address or set defaultAccount.")
        checksummed = toChecksumAddress(address)
        if checksummed not in self.accounts:
            raise ValueError("The address you specified is not added in your wallet. Try adding it first.")
        return checksummed

    async def getBalance(self, address: str = None) -> int:
        """
        Returns the AIN balance of the address.
        """
        if address is None:
            addr = self.getImpliedAddress()
        else:
            addr = toChecksumAddress(address)
        return await self.ain.db.ref(f"/accounts/{addr}/balance").getValue()

    async def transfer(self, toAddress: str, value: int, fromAddress: str = None, nonce: int = None):
        """
        Sends a transfer transaction to the network.
        """
        fromAddr = self.getImpliedAddress(fromAddress)
        toAddr = toChecksumAddress(toAddress)
        transferRef = self.ain.db.ref(f"/transfer/{fromAddr}/{toAddr}").push()
        return transferRef.setValue(
            ValueOnlyTransactionInput(
                ref="/value",
                address=fromAddress,
                value=value,
                nonce=nonce,
            )
        )

    def sign(self, data: str, address: str = None) -> str:
        """
        Signs a string data with the private key of the given address. It will use
        the defaultAccount if an address is not provided.
        """
        addr = self.getImpliedAddress(address)
        privateKeyBytes = bytes.fromhex(self.accounts[addr].private_key)
        return ecSignMessage(data, privateKeyBytes, self.chainId)
    
    def signTransaction(self, tx: TransactionBody, address: str = None) -> str:
        """
        Signs a transaction data with the private key of the given address. It will use
        the defaultAccount if an address is not provided.
        """
        addr = self.getImpliedAddress(address)
        privateKeyBytes = bytes.fromhex(self.accounts[addr].private_key)
        return ecSignTransaction(tx, privateKeyBytes, self.chainId)

    def getHashStrFromSig(self, signature: str) -> str:
        """
        Gets the hash from the signature.
        """
        sigBytes = toBytes(signature)
        lenHash = len(sigBytes) - 65
        hashedData = sigBytes[0:lenHash]
        return "0x" + hashedData.hex()

    def recover(self, signature: str) -> str:
        """
        Recovers an address of the account that was used to create the signature.
        """
        sigBytes = toBytes(signature)
        lenHash = len(sigBytes) - 65
        hashedData = sigBytes[0:lenHash]
        sig = ECDSASignature.fromSignature(sigBytes[lenHash:])
        return toChecksumAddress(bytesToHex(pubToAddress(ecRecoverPub(hashedData, sig, self.chainId)[1:])))

    def verifySignature(self, data: Any, signature: str, address: str) -> bool:
        return ecVerifySig(data, signature, address, self.chainId)

    # TODO(kriii): implement this function.
    def toV3Keystore(self):
        """
        Save the accounts in the wallet as V3 Keystores, locking them with the password.
        """
        pass
    
    # TODO(kriii): implement this function.
    def accountToV3Keystore(self):
        """
        Converts an account into a V3 Keystore and encrypts it with a password.
        """
        pass


import random
from ain.utils import getTimestamp
from typing import List

"""
SEE --> https://gist.github.com/mikelehen/3596a30bd69384624c11

Generator that creates 20-character string identifiers with the following properties:

1. They're based on timestamp so that they sort *after* any existing ids.
2. They contain 72-bits of random data after the timestamp so that IDs won't collide.
3. They sort *lexicographically* (timestamp converted to chars which will sort)
4. They're monotonically increasing. This is done by storing previous random bits
    increasing them only by one.
"""

# ASCII ordered base64 web safe chars.
ASCII_CHARS = "-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"

class PushId:
    """
    Timestamp of last push, used to prevent local
    collisions if you push twice in one ms.
    """
    _lastPushTime: int
    """
    We generate 72-bits of randomness which get
    turned into 12 characters and appended to the
    timestamp to prevent collisions with other clients.
    We store the last characters we generated because
    in the event of a collision, we'll use those same
    characters except "incremented" by one.
    """
    _lastRandChars: bytearray

    def __init__(self):
        self._lastPushTime = 0
        self._lastRandChars = bytearray(12)

    def generate(self) -> str:
        now = getTimestamp()
        duplicateTime = (now == self._lastPushTime)
        self._lastPushTime = now
        timeStampChars: List[str] = []
        for i in range(8):
            timeStampChars.append(ASCII_CHARS[now % 64])
            now //= 64
        if now != 0:
            raise ValueError("We should have converted the entire timestamp.")
        id = "".join(reversed(timeStampChars))
        if not duplicateTime:
            for i in range(12):
                self._lastRandChars[i] = random.randint(0, 63)
        else:
            i = 11
            while i >= 0 and self._lastRandChars[i] == 63:
                self._lastRandChars[i] = 0
                i -= 1
            if i >= 0:
                self._lastRandChars[i] += 1
        for i in range(12):
            id += ASCII_CHARS[self._lastRandChars[i]]
        if len(id) != 20:
            raise AssertionError("Length should be 20.")
        return id

    # TODO(kriii): Implement `Timestamp`.


from typing import Any, Literal, Optional, List, Union

class ECDSASignature:
    r: bytes
    s: bytes
    v: int

    def __init__(self, r: bytes, s: bytes, v: int):
        if len(r) != 32:
            raise ValueError("Invalid signature length: r")
        if len(s) != 32:
            raise ValueError("Invalid signature length: s")
        if v < 0 or v > 255:
            raise ValueError("Invalid signature: v")

        self.r = r
        self.s = s
        self.v = v

    @classmethod
    def fromSignature(cls, signature: bytes):
        """
        Converts signature format of the `eth_sign` RPC method to signature parameters.
        """
        if len(signature) != 65:
            raise ValueError("Invalid signature length: r")

        r = signature[0:32]
        s = signature[32:64]
        v = signature[64]
        return cls(r, s, v)

SetMultiOperationType = Literal["SET"]

GetMultiOperationType = Literal["GET"]

SetOperationType = Literal["SET_VALUE", "INC_VALUE", "DEC_VALUE", "SET_RULE", "SET_OWNER", "SET_FUNCTION"]

GetOperationType = Literal["GET_VALUE", "GET_RULE", "GET_OWNER", "GET_FUNCTION"]

OwnerPermission = Literal["branch_owner", "write_function", "write_owner", "write_rule"]

class SetOperation():
    type: SetOperationType
    ref: str
    value: Optional[Any]
    is_global: Optional[bool]

    def __init__(
        self,
        type: SetOperationType,
        ref: str,
        value: Any = None,
        is_global: bool = None,
    ):
        self.type = type
        self.ref = ref
        if value is not None:
            self.value = value
        if is_global is not None:
            self.is_global = is_global

class SetMultiOperation():
    type: SetMultiOperationType
    op_list: List[SetOperation]

    def __init__(
        self,
        type: SetMultiOperationType,
        op_list: List[SetOperation]
    ):
        self.type = type
        self.op_list = op_list

class GetOperation():
    type: GetOperationType
    ref: Optional[str]
    is_global: Optional[bool]

    def __init__(
        self,
        type: GetOperationType,
        ref: str = None,
        is_global: bool = None,
    ):
        self.type = type
        if ref is not None:
            self.ref = ref
        if is_global is not None:
            self.is_global = is_global

class GetMultiOperation():
    type: GetMultiOperationType
    op_list: List[GetOperation]

    def __init__(
        self,
        type: GetMultiOperationType,
        op_list: List[GetOperation]
    ):
        self.type = type
        self.op_list = op_list

class TransactionBodyBase:
    parent_tx_hash: Optional[str]
    operation: Union[SetOperation, SetMultiOperation]

class ValueOnlyTransactionBodyBase:
    parent_tx_hash: Optional[str]
    value: Optional[Any]
    ref: Optional[str]
    is_global: Optional[bool]

class TransactionInputBase:
    nonce: Optional[int]
    address: Optional[str]
    timestamp: Optional[int]

class TransactionBody(TransactionBodyBase):
    nonce: int
    timestamp: int
    def __init__(
        self,
        operation: Union[SetOperation, SetMultiOperation],
        nonce: int,
        timestamp: int,
        parent_tx_hash: str = None,
    ):
        self.operation = operation
        self.nonce = nonce
        self.timestamp = timestamp
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash


class TransactionInput(TransactionBodyBase, TransactionInputBase):
    def __init__(
        self,
        operation: Union[SetOperation, SetMultiOperation],
        parent_tx_hash: str = None,
        nonce: int = None,
        address: str = None,
        timestamp: int = None,
    ):
        self.operation = operation
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash
        if nonce is not None:
            self.nonce = nonce
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp

class ValueOnlyTransactionInput(ValueOnlyTransactionBodyBase, TransactionInputBase):
    def __init__(
        self,
        parent_tx_hash: str = None,
        value: Any = None,
        ref: str = None,
        is_global: bool = None,
        nonce: int = None,
        address: str = None,
        timestamp: int = None,
    ):
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash
        if value is not None:
            self.value = value
        if ref is not None:
            self.ref = ref
        if is_global is not None:
            self.is_global = is_global
        if nonce is not None:
            self.nonce = nonce
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp

class SetMultiTransactionInput(TransactionInputBase):
    parent_tx_hash: Optional[str]
    op_list: List[SetOperation]

    def __init__(
        self,
        op_list: List[SetOperation],
        parent_tx_hash: str = None,
        nonce: int = None,
        address: str = None,
        timestamp: int = None,
    ):
        self.op_list = op_list
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash
        if nonce is not None:
            self.nonce = nonce
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp

class Transaction():
    tx_body: TransactionBody
    signature: str
    hash: str
    address: str

class EvalRuleInput:
    value: Any
    ref: Optional[str]
    address: Optional[str]
    timestamp: Optional[int]
    is_global: Optional[bool]

class EvalOwnerInput:
    ref: Optional[str]
    address: Optional[str]
    permission: OwnerPermission
    is_global: Optional[bool]

class MatchInput:
    ref: Optional[str]
    is_global: Optional[bool]



from typing import TYPE_CHECKING
from ain.ain_db.ref import Reference

if TYPE_CHECKING:
    from ain.ain import Ain
    from ain.provider import Provider

class Database:
    _ain: "Ain"
    _provider: "Provider"
    
    def __init__(self, ain: "Ain", provider: "Provider"):
        self._ain = ain
        self._provider = provider

    def ref(self, path: str = None) -> Reference:
        return Reference(self._ain, path)




import aiohttp
import asyncio
import json
import inspect
from urllib.parse import urlparse, urljoin
from jsonrpcclient import request, parse, Ok, Error
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ain.ain import Ain

JSON_RPC_ENDPOINT = "json-rpc"

class Provider:
    endPoint: str
    apiEndPoint: str
    _ain: "Ain"
    
    def __init__(self, ain: "Ain", endpoint: str):
        self._ain = ain
        parsed = urlparse(endpoint)
        self.endPoint = parsed.geturl()
        self.apiEndPoint = urljoin(self.endPoint, JSON_RPC_ENDPOINT)

    async def send(self, rpcMethod: str, params: dict = {}) -> Any:
        params.update({"protoVer": self._ain.net.protoVer})
        dump = json.dumps(
            request(rpcMethod, params=params, id=0),
            default=lambda o: o.__dict__,
            separators=(",", ":"),
            sort_keys=True
        )
        async with aiohttp.ClientSession() as session:
            async with session.post(self.apiEndPoint, json=json.loads(dump)) as response:
                res = await response.read()
                parsed = parse(json.loads(res))
                if isinstance(parsed, Ok):
                    return parsed.result.get("result", None)
                if isinstance(parsed, Error):
                    return parsed
                raise AssertionError("The response must be singular.")



from ain.provider import Provider
from ain.constants import BLOCKCHAIN_PROTOCOL_VERSION

class Network:
    provider: Provider
    protoVer: str
    
    def __init__(self, provider: Provider):
        self.provider = provider
        self.protoVer = BLOCKCHAIN_PROTOCOL_VERSION



import re
import json
import time
from typing import Any, Union
from Crypto.Hash import keccak as _keccak
from coincurve import PrivateKey, PublicKey
from coincurve.ecdsa import deserialize_recoverable, recoverable_convert, cdata_to_der
from coincurve.utils import verify_signature, validate_secret
from ain.types import ECDSASignature, TransactionBody

def encodeVarInt(number: int) -> bytes:
    """
    Encodes the number as bitcoin variable length integer
    """
    MAX_SAFE_INTEGER = 9007199254740991
    if number < 0 or number > MAX_SAFE_INTEGER:
        raise ValueError("Number out of range")

    if number < 0xfd:
        byte = 1
    elif number <= 0xffff:
        number = (number << 8) + 0xfd
        byte = 3
    elif number <= 0xffffffff:
        number = (number << 8) + 0xfe
        byte = 5
    else:
        number = (number << 8) + 0xff
        byte = 9

    return number.to_bytes(byte, "little")

SIGNED_MESSAGE_PREFIX = "AINetwork Signed Message:\n"
SIGNED_MESSAGE_PREFIX_BYTES = bytes(SIGNED_MESSAGE_PREFIX, "utf-8")
SIGNED_MESSAGE_PREFIX_LENGTH = encodeVarInt(len(SIGNED_MESSAGE_PREFIX))

def getTimestamp() -> int:
    """
    Returns the current timestamp.
    """
    return int(time.time() * 1000)

def addHexPrefix(string: str) -> str:
    """
    Adds '0x' to the given string if it does not already start with '0x'.
    """
    return string if isHexPrefixed(string) else ("0x" + string)

def bytesToHex(input: bytes) -> str:
    """
    Converts the given bytes into a hex prefixed hex string.
    """
    return "0x" + input.hex()

def isValidAddress(string: str) -> bool:
    """
    Checks if the given string is a valid address.
    """
    pattern = re.compile("0x[0-9a-fA-F]{40}")
    return pattern.fullmatch(string) is not None

def isHexString(string: str) -> bool:
    """
    Checks if the given string is a hex string.
    """
    pattern = re.compile("0x[0-9a-fA-F]*")
    return pattern.fullmatch(string) is not None

def isHexPrefixed(string: str) -> bool:
    """
    Checks if the given string is prefixed with `0x`.
    """
    return string.startswith("0x")

def stripHexPrefix(string: str) -> str:
    """
    Removes `0x` from the given string.
    """
    return string[2:] if isHexPrefixed(string) else string

def padToEven(string: str) -> str:
    """
    Pads the given string to have an even length.
    """
    return string if len(string) % 2 == 0 else ("0" + string)

def toBytes(input: Any) -> bytes:
    """
    Attempts to turn a value into a bytes,
    As input it supports bytes, list(of ints), string, int, None.
    This function is equivalent to `toBuffer` from `@ainblockchain/ain-util`.
    """
    if type(input) is bytes:
        pass
    elif type(input) is list:
        input = bytes(input)
    elif type(input) is str:
        if isHexString(input):
            input = bytes.fromhex(padToEven(stripHexPrefix(input)))
        else:
            input = bytes(input, "utf-8")
    elif type(input) is int:
        input = bytes.fromhex(padToEven(stripHexPrefix(hex(input))))
    elif input is None:
        input = bytes(0)
    else:
        raise TypeError("Invalid type")
    return input

def toChecksumAddress(address: str) -> str:
    """
    Returns a checksummed address.
    """
    if not isValidAddress(address):
        raise ValueError("Invalid address")

    address = stripHexPrefix(address).lower()
    hash = keccak(address).hex()
    ret = "0x"

    for i in range(len(address)):
        if int(hash[i], base=16) >= 8:
            ret += address[i].upper()
        else:
            ret += address[i]

    return ret

def keccak(input: Any, bits: int = 256) -> bytes:
    """
    Creates Keccak hash of the input.
    """
    input = toBytes(input)
    k = _keccak.new(digest_bits=bits)
    k.update(input)
    return k.digest()

def hashTransaction(transaction: Union[TransactionBody, str]):
    """
    Generates keccak hash using a transaction body.
    """
    if type(transaction) is TransactionBody:
        transaction = json.dumps(
            transaction.__dict__,
            default=lambda o: o.__dict__,
            separators=(",", ":"),
            sort_keys=True
        )
    return keccak(keccak(transaction))

def hashMessage(message: Any):
    """
    Returns the bitcoin's varint encoding of keccak-256 hash of `message`,
    prefixed with the header `AINetwork Signed Message:`.
    """
    msgBytes = toBytes(message)
    msgLenBytes = encodeVarInt(len(msgBytes))
    dataBytes = bytes(
        bytearray(SIGNED_MESSAGE_PREFIX_LENGTH)
        + bytearray(SIGNED_MESSAGE_PREFIX_BYTES)
        + bytearray(msgLenBytes)
        + bytearray(msgBytes)
    )
    return keccak(keccak(dataBytes))

def ecSignHash(msgHash: bytes, privateKey: bytes, chainId: int = None) -> ECDSASignature:
    """
    Returns the ECDSA signature of a message hash.
    """
    pbK = PrivateKey(privateKey)
    sig = pbK.sign_recoverable(msgHash, hasher=None)
    r = sig[0:32]
    s = sig[32:64]
    v = sig[64] + (27 if chainId is None else chainId * 2 + 35)
    return ECDSASignature(r, s, v)

def ecSignMessage(message: Any, privateKey: bytes, chainId: int = None) -> str:
    """
    igns a message with a private key and returns a string signature.
    """
    hashedMsg = hashMessage(message)
    signature = ecSignHash(hashedMsg, privateKey, chainId)
    return "0x" + hashedMsg.hex() + signature.r.hex() + signature.s.hex() + toBytes(signature.v).hex()

def ecSignTransaction(txData: TransactionBody, privateKey: bytes, chainId: int = None) -> str:
    """
    Signs a transaction body with a private key and returns a string signature.
    """
    hashedTx = hashTransaction(txData)
    signature = ecSignHash(hashedTx, privateKey, chainId)
    return "0x" + hashedTx.hex() + signature.r.hex() + signature.s.hex() + toBytes(signature.v).hex()

def ecRecoverPub(msgHash: bytes, signature: ECDSASignature, chainId: int = None) -> bytes:
    """
    ECDSA public key recovery from signature.
    Returns Recovered public key.
    """
    recovery = signature.v - (27 if chainId is None else chainId * 2 + 35)
    if recovery != 0 and recovery != 1:
        raise ValueError("Invalid signature v value")
    cSig = bytes(bytearray(signature.r) + bytearray(signature.s) + bytearray([recovery]))
    senderPubKey = PublicKey.from_signature_and_message(cSig, msgHash, hasher=None)
    nSig = cdata_to_der(recoverable_convert(deserialize_recoverable(cSig)))
    if not senderPubKey.verify(nSig, msgHash, hasher=None):
        raise ValueError('Public key verify failed')
    return senderPubKey.format(False)

def ecVerifySig(data: Any, signature: str, address: str, chainId: int = None):
    """
    Checks if the signature is valid.
    """
    sigBytes = toBytes(signature)
    lenHash = len(sigBytes) - 65
    hashedData = sigBytes[:lenHash]
    if type(data) is TransactionBody:
        if hashedData != hashTransaction(data):
            return False
    else:
        if hashedData != hashMessage(data):
            return False

    sig = ECDSASignature.fromSignature(sigBytes[lenHash:])
    try:
        pub = ecRecoverPub(hashedData, sig, chainId)
    except ValueError:
        return False

    addr = bytesToHex(pubToAddress(pub[1:]))
    return areSameAddresses(address, addr)

def isValidPrivate(privateKey: bytes) -> bool:
    """
    Checks whether the `privateKey` is a valid private key (follows the rules of the
    curve secp256k1).
    """
    try:
        if len(privateKey) != 32:
            return False
        validate_secret(privateKey)
        return True
    except:
        return False

# TODO(kriii): implement this function.
def isValidPublic(publicKey: bytes, isSEC1: bool = False):
    """
    Checks whether the `publicKey` is a valid public key (follows the rules of the
    curve secp256k1 and meets the AIN requirements).

    args:
        publicKey (bytes): The two points of an uncompressed key, unless sanitize is enabled
        isSEC1 (bool): Accept public keys in other formats
    """
    pass

def areSameAddresses(address1: str, address2: str) -> bool:
    """
    Checks if the two addresses are the same.
    """
    return toChecksumAddress(address1) == toChecksumAddress(address2)

def privateToPublic(privateKey: bytes) -> bytes:
    """
    Returns the public key of a given private key.
    
    args:
        privateKey (bytes): A private key must be 256 bits wide
    """
    prK = PrivateKey(privateKey)
    return prK.public_key.format(False)[1:]

# TODO(kriii): Support `isSEC1`.
def pubToAddress(publicKey: Union[bytes, str], isSEC1: bool = False) -> bytes:
    """
    Returns the AI Network address of a given public key.

    args:
        publicKey (Union[bytes, str]):
            AIN public key | SEC1 encoded public key
            When the type is `str`, it must be hex prefixed
        isSEC1 (bool) : Key is SEC1 encoded

    returns:
        bytes: lower 160 bits of the hash of `publicKey`
    """
    if type(publicKey) is str:
        publicKeyBytes = toBytes(publicKey)
    elif type(publicKey) is bytes:
        publicKeyBytes = publicKey
    else:
        raise ValueError("Invalid public key type")

    if len(publicKeyBytes) != 64:
        raise ValueError("Invalid public key")
    return keccak(publicKeyBytes)[-20:]

def privateToAddress(privateKey: bytes) -> str:
    return toChecksumAddress(bytesToHex(pubToAddress(privateToPublic(privateKey))))

# TODO(kriii): implement this function.
def encryptWithPublicKey():
    pass

# TODO(kriii): implement this function.
def decryptWithPrivateKey():
    pass

# TODO(kriii): implement this function.
def createAccount():
    pass

# TODO(kriii): implement this function.
def privateToV3Keystore():
    pass

# TODO(kriii): implement this function.
def v3KeystoreToPrivate():
    pass


from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

requirements = [
    "pycryptodome",
    "coincurve",
    "aiohttp[speedups]",
    "jsonrpcclient",
]

test_requirements = []

setup(
    author="AIN Dev Team",
    author_email="dev@ainetwork.ai",
    python_requires=">3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="AI Network Client Library for Python3",
    install_requires=requirements,
    license="MPL license",
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=["ain", "ainetwork", "ainblockchain", "API"],
    name="ain-py",
    packages=find_packages(include=["ain", "ain.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/ainblockchain/ain-py",
    version="0.1.0",
    zip_safe=False,
)



from typing import List, Union
from ain.provider import Provider
from ain.net import Network
from ain.wallet import Wallet
from ain.types import TransactionInput, TransactionBody
from ain.ain_db.db import Database
from ain.utils import getTimestamp, toChecksumAddress

class Ain:
    provider: Provider
    chainId: int
    db: Database
    net: Network
    wallet: Wallet

    def __init__(self, providerUrl: str, chainId: int = 0):
        self.provider = Provider(self, providerUrl)
        self.chainId = chainId
        self.net = Network(self.provider)
        self.wallet = Wallet(self, self.chainId)
        self.db = Database(self, self.provider)

    def setProvider(self, providerUrl: str, chainId: int = 0):
        """
        Sets a new provider
        """
        self.provider = Provider(self, providerUrl)
        self.chainId = chainId
        self.net = Network(self.provider)
        self.wallet = Wallet(self, self.chainId)
        self.db = Database(self, self.provider)

    # TODO(kriii): Return objects typing.

    async def getBlock(
        self,
        blockHashOrBlockNumber: Union[str, int],
        returnTransactionObjects: bool = False,
    ):
        """
        A coroutine returns a block with the given hash or block number.

        args:
            blockHashOrBlockNumber (Union[str, int]):
                Block hash or block number
            returnTransactionObjects (bool):
                If true, returns the full transaction objects;
                otherwise, returns only the transaction hashes
        """
        if type(blockHashOrBlockNumber) is str:
            return await self.provider.send(
                "ain_getBlockByHash",
                {
                    "getFullTransactions": returnTransactionObjects,
                    "hash": blockHashOrBlockNumber,
                },
            )
        elif type(blockHashOrBlockNumber) is int:
            return await self.provider.send(
                "ain_getBlockByNumber",
                {
                    "getFullTransactions": returnTransactionObjects,
                    "number": blockHashOrBlockNumber,
                },
            )
        else:
            raise TypeError("blockHashOrBlockNumber has invalid type")

    async def getProposer(
        self,
        blockHashOrBlockNumber: Union[str, int],
    ) -> str:
        """
        A coroutine returns the address of the forger of given block
        """
        if type(blockHashOrBlockNumber) is str:
            return await self.provider.send(
                "ain_getProposerByHash", {"hash": blockHashOrBlockNumber}
            )
        elif type(blockHashOrBlockNumber) is int:
            return await self.provider.send(
                "ain_getProposerByNumber", {"number": blockHashOrBlockNumber}
            )
        else:
            raise TypeError("blockHashOrBlockNumber has invalid type")

    async def getValidators(
        self,
        blockHashOrBlockNumber: Union[str, int],
    ):
        """
        A coroutine returns the list of validators for a given block
        """
        if type(blockHashOrBlockNumber) is str:
            return await self.provider.send(
                "ain_getValidatorsByHash", {"hash": blockHashOrBlockNumber}
            )
        elif type(blockHashOrBlockNumber) is int:
            return await self.provider.send(
                "ain_getValidatorsByNumber", {"number": blockHashOrBlockNumber}
            )
        else:
            raise TypeError("blockHashOrBlockNumber has invalid type")

    async def getTransaction(self, transactionHash: str):
        """
        Returns the transaction with the given transaction hash.
        """
        return await self.provider.send(
            "ain_getTransactionByHash", {"hash": transactionHash}
        )

    async def getStateUsage(self, appName: str):
        """
        Returns the transaction with the given transaction hash.
        """
        return await self.provider.send(
            "ain_getStateUsage", {"app_name": appName}
        )

    async def sendTransaction(self, transactionObject: TransactionInput):
        """
        Signs and sends a transaction to the network
        """
        txBody = await self.buildTransactionBody(transactionObject)
        signature = self.wallet.signTransaction(
            txBody, getattr(transactionObject, "address", None)
        )
        return await self.sendSignedTransaction(signature, txBody)

    async def sendSignedTransaction(self, signature: str, txBody: TransactionBody):
        """
        Sends a signed transaction to the network
        """
        return await self.provider.send(
            "ain_sendSignedTransaction", {"signature": signature, "tx_body": txBody}
        )

    # TODO(kriii): implement this function.
    async def sendTransactionBatch(self):
        """
        Sends signed transactions to the network.
        """
        pass

    # TODO(kriii): implement this function.
    def depositConsensusStake(self):
        """
        Sends a transaction that deposits AIN for consensus staking.
        """
        pass

    # TODO(kriii): implement this function.
    def withdrawConsensusStake(self):
        """
        Sends a transaction that withdraws AIN for consensus staking.
        """
        pass

    async def getConsensusStakeAmount(self, account: str = None) -> int:
        """
        Gets the amount of AIN currently staked for participating in consensus protocol.
        """
        if account is None:
            address = self.wallet.getImpliedAddress()
        else:
            address = toChecksumAddress(account)
        return await self.db.ref(f"/deposit_accounts/consensus/{address}").getValue()

    async def getNonce(self, args: dict) -> int:
        """
        Returns the current transaction count of account, which is the nonce of the account.

        Args:
            args: May contain a string 'address' and a string 'from' values.
                The 'address' indicates the address of the account to get the
                nonce of, and the 'from' indicates where to get the nonce from.
                It could be either the pending transaction pool ("pending") or
                the committed blocks ("committed"). The default value is "committed".
        """
        params = dict(args)
        if "address" in args:
            params["address"] = toChecksumAddress(args["address"])
        else:
            params["address"] = self.wallet.getImpliedAddress()

        if "from" in args:
            if args["from"] != "pending" and args["from"] != "committed":
                raise ValueError("'from' should be either 'pending' or 'committed'")

        ret = await self.provider.send("ain_getNonce", params)
        return ret

    async def buildTransactionBody(
        self, transactionInput: TransactionInput
    ) -> TransactionBody:
        """
        Builds a transaction body from transaction input.
        """
        address = self.wallet.getImpliedAddress(
            getattr(transactionInput, "address", None)
        )
        operation = transactionInput.operation
        parent_tx_hash = getattr(transactionInput, "parent_tx_hash", None)
        nonce = getattr(transactionInput, "nonce", None)
        if nonce is None:
            nonce = await self.getNonce({"address": address, "from": "pending"})
        timestamp = getattr(transactionInput, "timestamp", getTimestamp())

        return TransactionBody(
            operation=operation,
            parent_tx_hash=parent_tx_hash,
            nonce=nonce,
            timestamp=timestamp,
        )

    # TODO(kriii): implement this function.
    def __stakeFunction(self):
        """
        A base function for all staking related database changes. It builds a
        deposit/withdraw transaction and sends the transaction by calling sendTransaction().
        """
        pass